# Reflective, Refractive Leapfrog schemes
import copy
import numpy as np
from .Logposterior import Logpost
from .treeManipulation import idx2nodeMAP, NNI



def reflection(propB, propM, idx2node, stepsz, include=False):
    tmpB = propB + stepsz*propM
    ref_time = 0
    NNI_attempts, Ref_attempts = 0, 0
    while min(tmpB) <= 0:
        timelist = tmpB/abs(propM)
        ref_index = np.argmin(timelist)
        propB = propB + (stepsz-ref_time+timelist[ref_index]) * propM
        
        # reflect the momentum
        propM[ref_index] *= -1
        Ref_attempts += 1
        # perform NNI
        if not idx2node[ref_index].is_leaf():
            NNI_attempts += NNI(idx2node[ref_index], include=include)
        
        ref_time = stepsz + timelist[ref_index]
        tmpB = propB + (stepsz-ref_time) * propM
    
    return tmpB, propM, NNI_attempts, Ref_attempts


def refraction(prop_tree, propB, propM, D, U, U_inv, pden, L, idx2node, stepsz, 
               surrogate=True, include=False, scale = 0.1, delta = 0.01):
    tmpB = propB + stepsz*propM
    ref_time = 0
    NNI_attempts, Ref_attempts = 0, 0
    while min(tmpB) <= 0:
        timelist = tmpB/abs(propM)
        ref_index = np.argmin(timelist)
        propB = propB + (stepsz-ref_time+timelist[ref_index]) * propM
        
        propM[ref_index] *= -1
        Ref_attempts += 1
        if not idx2node[ref_index].is_leaf():
            if surrogate:
                U_before_nni = Logpost(prop_tree, propB, D, U, U_inv, pden, L, scale=scale, delta=delta, surrogate=True)
            
                tmp_tree = copy.deepcopy(prop_tree)
                tmp_idx2node = idx2nodeMAP(tmp_tree)
                # check if a topology transition has been made
                tmp_nni_made = NNI(tmp_idx2node[ref_index], include=include)
                
                if tmp_nni_made !=0:            
                    U_after_nni = Logpost(tmp_tree, propB, D, U, U_inv, pden, L, scale=scale, delta=delta, surrogate=True)
                    # compute the energy gap
                    delta_U = U_after_nni - U_before_nni
                    
                    if propM[ref_index]**2 >= 2*delta_U:
                        propM[ref_index] = np.sqrt(propM[ref_index]**2-2*delta_U)
                        prop_tree = tmp_tree
                        NNI_attempts += 1
            else:
                NNI_attempts += NNI(idx2node[ref_index], include=include)
        
        ref_time = stepsz + timelist[ref_index]
        tmpB = propB + (stepsz-ref_time) * propM
    
    return prop_tree, tmpB, propM, NNI_attempts, Ref_attempts




# Substitution Models

import numpy as np



def decompJC():
    pA = pG = pC = pT = .25
    rate_matrix_JC = 1/3.0 * np.ones((4,4))
    for i in range(4):
        rate_matrix_JC[i,i] = -1.0
    
    D_JC, U_JC = np.linalg.eig(rate_matrix_JC)
    U_JC_inv = np.linalg.inv(U_JC)
    
    return D_JC, U_JC, U_JC_inv, rate_matrix_JC


def decompHKY(pden, kappa):
    pA, pG, pC, pT = pden
    beta = 1.0/(2*(pA+pG)*(pC+pT) + 2*kappa*(pA*pG+pC*pT))
    rate_matrix_HKY = np.zeros((4,4))
    for i in range(4):
        for j in range(4):
            if j!=i:
                rate_matrix_HKY[i,j] = pden[j]
            if i+j == 1 or i+j == 5:
                rate_matrix_HKY[i,j] *= kappa
    
    for i in range(4):
        rate_matrix_HKY[i,i] = - sum(rate_matrix_HKY[i,])
    
    rate_matrix_HKY = beta * rate_matrix_HKY
    D_HKY, U_HKY = np.linalg.eig(rate_matrix_HKY)
    U_HKY_inv = np.linalg.inv(U_HKY)
       
    return D_HKY, U_HKY, U_HKY_inv, rate_matrix_HKY


def decompGTR(pden, AG, AC, AT, GC, GT, CT):
    pA, pG, pC, pT = pden
    beta = 1.0/(2*(AG*pA*pG+AC*pA*pC+AT*pA*pT+GC*pG*pC+GT*pG*pT+CT*pC*pT))
    rate_matrix_GTR = np.zeros((4,4))
    for i in range(4):
        for j in range(4):
            if j!=i:
                rate_matrix_GTR[i,j] = pden[j]
                if i+j == 1:
                    rate_matrix_GTR[i,j] *= AG
                if i+j == 2:
                    rate_matrix_GTR[i,j] *= AC
                if i+j == 3 and abs(i-j) > 1:
                    rate_matrix_GTR[i,j] *= AT
                if i+j == 3 and abs(i-j) == 1:
                    rate_matrix_GTR[i,j] *= GC
                if i+j == 4:
                    rate_matrix_GTR[i,j] *= GT
                if i+j == 5:
                    rate_matrix_GTR[i,j] *= CT
    
    for i in range(4):
        rate_matrix_GTR[i,i] = - sum(rate_matrix_GTR[i,])
    
    rate_matrix_GTR = beta * rate_matrix_GTR
    D_GTR, U_GTR = np.linalg.eig(rate_matrix_GTR)
    U_GTR_inv = np.linalg.inv(U_GTR)
    
    return D_GTR, U_GTR, U_GTR_inv, rate_matrix_GTR



# Logprior and its gradient
import numpy as np


# exponential prior for the branch lengths
def phyloLogprior_exp(branch, scale=0.1, grad=False):
    if not grad:
        return -np.sum(branch)/scale - np.log(scale)*len(branch)
    else:
        return -np.ones(len(branch))/scale


from ete3 import Tree
import numpy as np
import rateMatrix as rateM
import Loglikelihood
import Logprior
import Logposterior
import phyloHMC as phmc
import dataManipulation as data
import branchManipulation as branch
import treeManipulation as tree
import resultAnalysis as result 

